# popup.tcl --

# A popup widget.

# Version   : 0.0.1
# Author    : Mark G. Saye
# Email     : markgsaye@yahoo.com
# Copyright : Copyright (C) 2004
# Date      : March 17, 2004

# See the file "LICENSE.txt" or "LICENSE.html" for information on usage
# and distribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# ==================================================================== #

  namespace eval popup {}

# ==================================================================== #

# cget --

# Description
#   Get the value of a configuration option

# Arguments
#   W      : popup widget
#   option : name of configuration option

# Return
#   value of configuration option for widget W

proc popup::cget {W option} {

  debug 2 "proc cget \[$W\] \[$option\]\n"

# -------------------------------------------------------------------- #

  upvar #0 [.]::$W _

# -------------------------------------------------------------------- #

  switch -- $option {
    -focus -
    -grab {
      set return $_($option)
    }
    default {
      set return [$W cget $option]
    }
  }

# -------------------------------------------------------------------- #

  debug 2 "  cget \[$W\] \[$option\] return \[$return\]\n"
  return $return

}

# ==================================================================== #

# configure --

# Description
#   Configure a popup widget

# Arguments
#   args : list of option/value pairs

proc popup::configure {W args} {

  debug 2 "proc configure \[$W\] \[$args\]\n"

# -------------------------------------------------------------------- #

  variable map
  variable options
  variable options_toplevel

  upvar #0 [.]::$W _

# -------------------------------------------------------------------- #

  set error 0

  switch -- [llength $args] {
    0 {
      set return [list]
      foreach option [array names map] {
        set synonym [lindex $map($option) 1]
        lappend return [list $option $synonym]
      }
#     foreach option [array names options_toplevel] {
#       set config [$W configure $option]
#       lappend return [lreplace $config 3 3 $options_toplevel($option)]
#     }
      foreach option $options {
        lappend return [option:get $W $option]
      }
      set return [lsort -index 0 $return]
    }
    1 {
      set option [lindex $args 0]
      if { [info exists map($option)] } {
        set widget [lindex $map($option) 0]
        set option [lindex $map($option) 1]
      }
#     if { [lsearch $options $option] > -1 } {
        set return [option:get $W $option]
#     } else {
#       set return [$W configure $option]
#     }
    }
    default {
      foreach {option value} $args {
#       if { [info exists map($option)] } {
#         set widget [lindex $map($option) 0]
#         set option [lindex $map($option) 1]
#       }
#       if { [info exists options($option)] } {
          option:set $W $option $value
#       } else {
#         $W configure $option $value
#       }
      }
      set return ""
    }
  }

# -------------------------------------------------------------------- #

  return $return

}

# ==================================================================== #

# init:options_ --

# Description
#   Initialize widget options

proc popup::init:options_ {} {

  debug 2 "proc init:options_\n"

# -------------------------------------------------------------------- #

  variable class

  variable create [list -focus -grab]

  variable map ; array set map {}

  array set def [list \
    -focus %W \
    -grab  %W \
  ]

  variable db ; array set db [list \
    -focus [list focus Focus] \
    -grab  [list grab  Grab] \
  ]

# -------------------------------------------------------------------- #

  # igmore some toplevel properties
  set igmore_toplevel [list -screen -use]

  # import toplevel properties
  variable options_toplevel [list]

  foreach config [::. configure] {
    switch -- [llength $config] {
      2 {
        set synonym [lindex $config 0]
        set option  [lindex $config 1]
        set map($synonym) [list toplevel $option]
      }
      5 {
        set option  [lindex $config 0]
        set default [lindex $config 3]
        if { [lsearch $igmore_toplevel $option] == -1 } {
          lappend options_toplevel $option
          set def($option) $default
        }
      }
    }
  }

  # override some toplevel properties
  array set def [list \
    -borderwidth 1 \
    -class $class \
    -cursor left_ptr \
    -relief raised \
  ]

# # remove some options
# foreach option $igmore_toplevel {
#   unset def($option)
# }

# -------------------------------------------------------------------- #

  # options which may only be supplied at creation time
  variable create_toplevel \
    [list -class -colormap -container -visual -use]

  # options which are required at creation time
  # this includes default options which have been overridden
  variable require_toplevel \
    [list -borderwidth -class -cursor -relief]

# -------------------------------------------------------------------- #

  variable options  ; set options  [list]
  variable defaults ; set defaults [list]

  foreach option [lsort [array names def]] {
    lappend options $option
    lappend defaults $def($option)
    debug 3 "  option  = \[$option\] default = \[$def($option)\]\n"
  }
  debug 3 "  options  = \[$options\]\n"
  debug 3 "  defaults = \[$defaults\]\n"

# -------------------------------------------------------------------- #

  return

}

# ==================================================================== #

# option:find --

# Description
#   Find the full name of a configuration option

# Arguments
#   option : name of configuration option

# Return
#   index of option in options list

proc popup::option:find {option} {

  debug 4 "proc option:find \[$option\]\n"

# -------------------------------------------------------------------- #

  variable map
  variable options

# -------------------------------------------------------------------- #

  if { [info exists map($option)] } {
    set option [lindex $map($option) 1]
  }

# -------------------------------------------------------------------- #

  set index [lsearch -exact $options $option]

  if { $index == -1 } {
    set found [list]
    set i 0
    foreach opt $options {
      if { [string match $option* $opt] } { lappend found $i }
      incr i
    }
    debug 3 "  found \[[llength $found]\] matches\n"

    if { [llength $found] == 1 } {
      set index [lindex $found 0]
    } else {
      set index -1
    }
#   switch -- [llength $found] {
#     0 { return -code error "unknown option \"$option\"" }
#     1 { set option [lindex $match 0] }
#     default { return -code error "ambiguous option \"$option\"" }
#   }
  }

# -------------------------------------------------------------------- #

  debug 3 "  option:find \[$option\] return \[$index\]\n"
  return $index

}

# ==================================================================== #

# option:get --

# Description
#   Get a configuration option

# Arguments
#   W      : popup widget
#   option : name of configuration option

# Return
#   value of configuration option

proc popup::option:get {W option} {

  debug 4 "proc option:get \[$W\] \[$option\]\n"

# -------------------------------------------------------------------- #

  variable db
  variable defaults
  variable map
  variable options
  variable options_toplevel

  upvar #0 [.]::$W _

# -------------------------------------------------------------------- #

  set i [option:find $option]
  debug 3 "  i = \[$i\]\n"

  if { $i == -1 } {
    return -code error "unknown option \"$option\""
  } else {
    set option  [lindex $options  $i]
    set default [lindex $defaults $i]
  }
# debug 3 "  option  = \[$option\]\n"
# debug 3 "  default = \[$default\]\n"

# -------------------------------------------------------------------- #

# set error 0

# debug 3 "  switching on \[$option\]\n"
  switch -- $option {
    -focus -
    -grab {
      set value $_($option)
      set default $W
      set dbname  [lindex $db($option) 0]
      set dbclass [lindex $db($option) 1]
    }
    default {
#     debug 3 "  got default switch\n"
#     debug 3 "  option = \[$option\]\n"
#     debug 3 "  options_toplevel = \[$options_toplevel\]\n"
      if { [lsearch $options_toplevel $option] > -1 } {
#       debug 3 "  got toplevel option\n"
        set config [$W configure $option]
        set dbname  [lindex $config 1]
        set dbclass [lindex $config 2]
        set value   [lindex $config 4]
      } else {
#       debug 3 "  return -code \[error\] \[TODO: unknown option \"$option\"\]\n"
        return -code error "TODO: unknown option \"$option\""
      }
    }
  }

  set return [list $option $dbname $dbclass $default $value]

# -------------------------------------------------------------------- #

# if { $error } {
#   set code error
#   set errorCode $::errorCode
#   debug 3 "  error in proc option:get\n"
#   debug 3 "  error     = \[$error\]\n"
#   debug 3 "  return    = \[$return\]\n"
#   debug 3 "  errorCode = \[$errorCode\]\n"
# } else {
#   set code ok
# }

# -------------------------------------------------------------------- #

  debug 5 "  option:get \[$W\] \[$option\] return \[$return\]\n"
  return $return

}

# ==================================================================== #

# option:set --

# Description
#   Set a configuration option

# Arguments
#   W      : popup widget
#   option : name  of configuration option
#   value  : value of configuration option

proc popup::option:set {W option value} {

  debug 2 "proc option:set \[$W\] \[$option\] \[$value\]\n"

# -------------------------------------------------------------------- #

  variable defaults
  variable options
  variable options_toplevel

  upvar #0 [.]::$W _

# -------------------------------------------------------------------- #

  set i [option:find $option]

  if { $i == -1 } {
    return -code error "unknown option \"$option\""
  } else {
    set option  [lindex $options  $i]
    set default [lindex $defaults $i]
  }
  debug 3 "  option  = \[$option\]\n"
  debug 3 "  default = \[$default\]\n"

# -------------------------------------------------------------------- #

  switch -- $option {
    -focus -
    -grab {
      set _($option) $value
      if { [winfo ismapped $W] } {
        unmap $W
        after idle [list [.]::map $W]
      }
    }
    default {
      if { [lsearch $options_toplevel $option] > -1 } {
        $W configure $option $value
      } else {
        return -code error "unknown option \"$option\""
      }
    }
  }

# -------------------------------------------------------------------- #

  debug 3 "  option:set return\n"
  return

}

# ==================================================================== #

  return

# ==================================================================== #

